"""Constants for hassfeld"""
BROWSE_CHILDREN = "BrowseDirectChildren"
BROWSE_METADATA = "BrowseMetadata"
CID_SEARCH_ARTISTS = "0/My Music/Search/TrackArtists"
CID_SEARCH_ALBUMS = "0/My Music/Search/Albums"
CID_SEARCH_COMPOSERS = "0/My Music/Search/Composers"
CID_SEARCH_ALLTRACKS = "0/My Music/Search/AllTracks"
DEFAULT_PORT_WEBSERVICE = 47365
DELAY_REQUEST_FAILURE_LONG_POLLING = 60
DELAY_FAST_UPDATE_CHECKS = 0.1
USER_AGENT_RAUMFELD = "RaumfeldControl/3.10 RaumfeldProtocol"
USER_AGENT_RAUMFELD_OIDS = ["0/RadioTime", "0/Tidal"]
MAX_RETRIES = 100
PLAY_MODE_NORMAL = "NORMAL"
PLAY_MODE_SHUFFLE = "SHUFFLE"
PLAY_MODE_REPEAT_ONE = "REPEAT_ONE"
PLAY_MODE_REPEAT_ALL = "REPEAT_ALL"
PLAY_MODE_RANDOM = "RANDOM"
POWER_ACTIVE = "active"
POWER_STANDBY_AUTOMATIC = "AUTOMATIC_STANDBY"
POWER_STANDBY_MANUAL = "MANUAL_STANDBY"
PREFERRED_TIMEOUT_LONG_POLLING = 300
REQUIRED_METADATA = {"DIDL-Lite": {"item": {"upnp:class": "object.item.audioItem"}}}
SPOTIFY_ACTIVE = "ACTIVE"
TRANSPORT_STATE_NO_MEDIA = "NO_MEDIA_PRESENT"
TRANSPORT_STATE_PAUSED = "PAUSED_PLAYBACK"
TRANSPORT_STATE_PLAYING = "PLAYING"
TRANSPORT_STATE_STOPPED = "STOPPED"
TRANSPORT_STATE_TRANSITIONING = "TRANSITIONING"
TIMEOUT_UPNP = 15
TIMEOUT_LONG_POLLING = 330
TIMEOUT_WEBSERVICE_ACTION = 5
TRIGGER_UPDATE_DEVICES = "devices"
TRIGGER_UPDATE_HOST_INFO = "host_info"
TRIGGER_UPDATE_SYSTEM_STATE = "system_state"
TRIGGER_UPDATE_ZONE_CONFIG = "zone_config"
TYPE_MEDIA_SERVER = "urn:schemas-upnp-org:device:MediaServer:1"
TYPE_RAUMFELD_DEVICE = "urn:schemas-raumfeld-com:device:RaumfeldDevice:1"
# TODO: replace CONTENT_DIRECTORY by SERVICE_CONTENT_DIRECTORY
CONTENT_DIRECTORY = "urn:schemas-upnp-org:service:ContentDirectory:1"
SERVICE_ID_SETUP_SERVICE = "urn:schemas-raumfeld-com:service:SetupService:1"
SERVICE_AV_TRANSPORT = "urn:schemas-upnp-org:service:AVTransport:1"
SERVICE_RENDERING_CONTROL = "urn:schemas-upnp-org:service:RenderingControl:1"
SERVICE_CONTENT_DIRECTORY = "urn:schemas-upnp-org:service:ContentDirectory:1"
SOUND_SUCCESS = "Success"
SOUND_FAILURE = "Failure"
RESPONSE_KEY_CURRENT_MUTE = "CurrentMute"
RESPONSE_KEY_CURRENT_VOLUME = "CurrentVolume"
RESPONSE_KEY_RESULT = "Result"
